from collections.abc import MutableMapping
from typing import Union

import numpy as np
from google.protobuf.pyext._message import MessageMapContainer

from .score import NamedScore
from ..base import BaseProtoView
from ..helper import typename
from ..proto.docarray_pb2 import NamedScoreProto


class NamedScoreMap(BaseProtoView, MutableMapping):
    """
    It offers a Pythonic interface to allow users access and manipulate
    :class:`docarray_pb2.NamedScoreMappingProto` object without working with Protobuf itself.

    It offers an interface to access and update scores as `NamedScore` as values of a `dict` with a string key.

    To create a :class:`NamedScoreMappingProto` object, simply:

        .. highlight:: python
        .. code-block:: python

            from jina.types.score.map import NamedScoreMapping
            scores = NamedScoreMapping()
            scores['score'] = 50

    :class:`NamedScoreMapping` can be built from ``docarray_pb2.NamedScoreMappingProto`` (as a weak reference or a deep copy)

    :param scores: The scores to construct from, depending on the ``copy``,
        it builds a view or a copy from it.
    :type score: Optional[docarray_pb2.NamedScoreMappingProto]
    :param copy: When ``scores`` is given as a :class:`NamedScoreMappingProto` object, build a
        view (i.e. weak reference) from it or a deep copy from it.
    :type copy: bool
    :param kwargs: Other parameters to be set

    """

    _PbMsg = MessageMapContainer

    def __setitem__(
        self,
        key: str,
        value: Union[NamedScoreProto, NamedScore, float, np.generic],
    ):
        if isinstance(value, NamedScoreProto):
            self._pb_body[key].CopyFrom(value)
        elif isinstance(value, NamedScore):
            self._pb_body[key].CopyFrom(value._pb_body)
        elif isinstance(value, (np.generic, float, int)):
            self._pb_body[key].value = float(value)
        else:
            raise TypeError(f'value is in unsupported type {typename(value)}')

    def __getitem__(
        self,
        key: str,
    ) -> 'NamedScore':
        return NamedScore(self._pb_body[key])

    def __delitem__(
        self,
        key: str,
    ):
        del self._pb_body[key]

    def __contains__(self, key: str):
        return key in self._pb_body

    def __iter__(self):
        for key in self._pb_body:
            yield key

    def __len__(self):
        return len(self._pb_body)



from typing import List, Sequence, TYPE_CHECKING, Optional, Union

from ...ndarray import NdArray

if TYPE_CHECKING:
    from ...document import ArrayType
    from ...document.mixins.content import DocumentContentType


class ContentPropertyMixin:
    """Helpers that provide faster getter & setter for :attr:`.content` by using protobuf directly. """

    def _check_length(self, target_len: int):
        if target_len != len(self):
            raise ValueError(
                f'Length must match {self!r}, but {target_len} != {len(self)}'
            )

    @property
    def embeddings(self) -> Optional['ArrayType']:
        """Return a :class:`ArrayType` stacking all the `embedding` attributes as rows.

        :return: a :class:`ArrayType` of embedding
        """
        if self:
            return NdArray.unravel([d.embedding for d in self._pb_body])

    @embeddings.setter
    def embeddings(self, value: 'ArrayType'):
        """Set the :attr:`.embedding` of the Documents.

        To remove all embeddings of all Documents:
        .. highlight:: python
        .. code-block:: python

            da.embeddings = None

        :param value: The embedding matrix to set
        """

        if value is None:
            for d in self._pb_body:
                d.ClearField('embedding')
        else:
            emb_shape0 = _get_len(value)
            self._check_length(emb_shape0)
            NdArray.ravel(value, self, 'embedding')

    @property
    def blobs(self) -> Optional['ArrayType']:
        """Return a :class:`ArrayType` stacking all :attr:`.blob`.

        The `blob` attributes are stacked together along a newly created first
        dimension (as if you would stack using ``np.stack(X, axis=0)``).

        .. warning:: This operation assumes all blobs have the same shape and dtype.
                 All dtype and shape values are assumed to be equal to the values of the
                 first element in the DocumentArray / DocumentArrayMemmap

        :return: a :class:`ArrayType` of blobs
        """
        if self:
            return NdArray.unravel([d.blob for d in self._pb_body])

    @blobs.setter
    def blobs(self, value: 'ArrayType'):
        """Set :attr:`.blob` of the Documents. To clear all :attr:`blob`, set it to ``None``.

        :param value: The blob array to set. The first axis is the "row" axis.
        """

        if value is None:
            for d in self._pb_body:
                d.ClearField('blob')
        else:
            blobs_shape0 = _get_len(value)
            self._check_length(blobs_shape0)

            NdArray.ravel(value, self, 'blob')

    @property
    def texts(self) -> Optional[List[str]]:
        """Get :attr:`.text` of all Documents

        :return: a list of texts
        """
        if self:
            return [d.text for d in self._pb_body]

    @texts.setter
    def texts(self, value: Sequence[str]):
        """Set :attr:`.text` for all Documents. To clear all :attr:`text`, set it to ``None``.

        :param value: A sequence of texts to set, should be the same length as the
            number of Documents
        """
        if value is None:
            for d in self._pb_body:
                d.ClearField('text')
        else:
            self._check_length(len(value))

            for doc, text in zip(self._pb_body, value):
                doc.text = text

    @property
    def buffers(self) -> Optional[List[bytes]]:
        """Get the buffer attribute of all Documents.

        :return: a list of buffers
        """
        if self:
            return [d.buffer for d in self._pb_body]

    @buffers.setter
    def buffers(self, value: List[bytes]):
        """Set the buffer attribute for all Documents. To clear all :attr:`buffer`, set it to ``None``.

        :param value: A sequence of buffer to set, should be the same length as the
            number of Documents
        """

        if value is None:
            for d in self._pb_body:
                d.ClearField('buffer')
        else:
            self._check_length(len(value))

            for doc, buffer in zip(self._pb_body, value):
                doc.buffer = buffer

    @property
    def contents(self) -> Optional[Union[Sequence['DocumentContentType'], 'ArrayType']]:
        """Get the :attr:`.content` of all Documents.

        :return: a list of texts, buffers or :class:`ArrayType`
        """
        if self:
            content_type = self[0].content_type
            if content_type:
                return getattr(self, f'{self[0].content_type}s')

    @contents.setter
    def contents(
        self, value: Sequence[Union[Sequence['DocumentContentType'], 'ArrayType']]
    ):
        """Set the :attr:`.content` of all Documents.

        :param value: a list of texts, buffers or :class:`ArrayType`
        """
        if self:
            content_type = self[0].content_type
            if content_type:
                setattr(self, f'{self[0].content_type}s', value)


def _get_len(value):
    return len(value) if isinstance(value, (list, tuple)) else value.shape[0]



import json
import mimetypes
import uuid
from typing import (
    Dict,
    Iterable,
    Optional,
    TypeVar,
    Union,
    overload,
    TYPE_CHECKING,
    Tuple,
    Sequence,
)

import numpy as np
from google.protobuf import json_format

from .mixins import AllMixins
from .mixins.version import versioned

from ..base import BaseProtoView
from ..helper import typename
from ..ndarray import NdArray
from ..proto.docarray_pb2 import DocumentProto
from ..simple import StructView, NamedScoreMap

if TYPE_CHECKING:
    from ..ndarray import ArrayType
    from ..simple import NamedScore
    from ..array.chunk import ChunkArray
    from ..array.match import MatchArray

    DocumentSourceType = TypeVar('DocumentSourceType', bytes, str, Dict)


_all_mime_types = set(mimetypes.types_map.values())
_all_doc_content_keys = {'content', 'blob', 'text', 'buffer'}


class Document(AllMixins, BaseProtoView):
    """
    :class:`Document` is one of the **primitive data type** in Jina.

    It offers a Pythonic interface to allow users access and manipulate
    :class:`jina.docarray_pb2.DocumentProto` object without working with Protobuf itself.

    To create a :class:`Document` object, simply:

        .. highlight:: python
        .. code-block:: python

            from jina import Document
            d = Document()
            d.text = 'abc'

    Jina requires each Document to have a string id. You can set a custom one,
    or if non has been set a random one will be assigned.

    To access and modify the content of the document, you can use :attr:`text`, :attr:`blob`, and :attr:`buffer`.
    Each property is implemented with proper setter, to improve the integrity and user experience. For example,
    assigning ``doc.blob`` or ``doc.embedding`` can be simply done via:

        .. highlight:: python
        .. code-block:: python

            import numpy as np

            # to set as content
            d.content = np.random.random([10, 5])

            # to set as embedding
            d.embedding = np.random.random([10, 5])

    MIME type is auto set/guessed when setting :attr:`content` and :attr:`uri`

    :class:`Document` also provides multiple way to build from existing Document. You can build :class:`Document`
    from ``docarray_pb2.DocumentProto``, ``bytes``, ``str``, and ``Dict``. You can also use it as view (i.e.
    weak reference when building from an existing ``docarray_pb2.DocumentProto``). For example,

        .. highlight:: python
        .. code-block:: python

            a = DocumentProto()
            b = Document(a, copy=False)
            a.text = 'hello'
            assert b.text == 'hello'

    You can leverage the :meth:`convert_a_to_b` interface to convert between content forms.

    """

    # overload_inject_start_document
    @overload
    def __init__(
        self,
        adjacency: Optional[int] = None,
        blob: Optional['ArrayType'] = None,
        buffer: Optional[bytes] = None,
        chunks: Optional[Iterable['Document']] = None,
        embedding: Optional['ArrayType'] = None,
        granularity: Optional[int] = None,
        id: Optional[str] = None,
        location: Optional[Sequence[float]] = None,
        matches: Optional[Iterable['Document']] = None,
        mime_type: Optional[str] = None,
        modality: Optional[str] = None,
        offset: Optional[float] = None,
        parent_id: Optional[str] = None,
        tags: Optional[Union[Dict, StructView]] = None,
        text: Optional[str] = None,
        uri: Optional[str] = None,
        weight: Optional[float] = None,
        **kwargs,
    ):
        """
        :param adjacency: the adjacency of this Document
        :param blob: the blob content of thi Document
        :param buffer: the buffer bytes from this document
        :param chunks: the array of chunks of this document
        :param embedding: the embedding of this Document
        :param granularity: the granularity of this Document
        :param id: the id of this Document
        :param location: location info in a tuple.
        :param matches: the array of matches attached to this document
        :param mime_type: the mime_type of this Document
        :param modality: the modality of the document.
        :param offset: the offset
        :param parent_id: the parent id of this Document
        :param tags: a Python dict view of the tags.
        :param text: the text from this document content
        :param uri: the uri of this Document
        :param weight: the weight of the document
        :param kwargs: other parameters to be set _after_ the document is constructed
        """

    # overload_inject_end_document

    _PbMsg = DocumentProto

    def __init__(
        self,
        obj: Optional[Union['DocumentSourceType', 'Document']] = None,
        field_resolver: Dict[str, str] = None,
        copy: bool = False,
        **kwargs,
    ):
        """
        :param obj: the document to construct from. If ``bytes`` is given
                then deserialize a :class:`DocumentProto`; ``dict`` is given then
                parse a :class:`DocumentProto` from it; ``str`` is given, then consider
                it as a JSON string and parse a :class:`DocumentProto` from it; finally,
                one can also give `DocumentProto` directly, then depending on the ``copy``,
                it builds a view or a copy from it.
        :param copy: when ``document`` is given as a :class:`DocumentProto` object, build a
                view (i.e. weak reference) from it or a deep copy from it.
        :param field_resolver: a map from field names defined in JSON, dict to the field
            names defined in Document.
        :param kwargs: other parameters to be set _after_ the document is constructed

        .. note::

            When ``document`` is a JSON string or Python dictionary object, the constructor will only map the values
            from known fields defined in Protobuf, all unknown fields are mapped to ``document.tags``. For example,

            .. highlight:: python
            .. code-block:: python

                d = Document({'id': '123', 'hello': 'world', 'tags': {'good': 'bye'}})

                assert d.id == '123'  # true
                assert d.tags['hello'] == 'world'  # true
                assert d.tags['good'] == 'bye'  # true
        """
        try:
            super().__init__(obj, copy=copy)

            if self._pb_body is None and isinstance(obj, (dict, str)):
                # note that not any dict can be parsed in this branch. As later we use Protobuf JSON parser, this
                # dict must be a valid JSON. The next lines make sure the dict is a valid json.
                if isinstance(obj, dict):
                    try:
                        json.dumps(obj)
                    except:
                        raise json_format.ParseError

                if isinstance(obj, str):
                    obj = json.loads(obj)

                if field_resolver:
                    obj = {field_resolver.get(k, k): v for k, v in obj.items()}

                user_fields = set(obj)
                support_fields = set(
                    self.attributes(
                        include_proto_fields_camelcase=True,
                        include_properties=False,
                    )
                )

                self._pb_body = self._PbMsg()
                if support_fields.issuperset(user_fields):
                    json_format.ParseDict(obj, self._pb_body)
                else:
                    _intersect = support_fields.intersection(user_fields)
                    _remainder = user_fields.difference(_intersect)
                    if _intersect:
                        json_format.ParseDict(
                            {k: obj[k] for k in _intersect}, self._pb_body
                        )
                    if _remainder:
                        support_prop = set(
                            self.attributes(
                                include_proto_fields=False, include_properties=True
                            )
                        )
                        _intersect2 = support_prop.intersection(_remainder)
                        _remainder2 = _remainder.difference(_intersect2)

                        if _intersect2:
                            self._set_attributes(**{p: obj[p] for p in _intersect2})

                        if _remainder2:
                            self._pb_body.tags.update({k: obj[k] for k in _remainder})

            if self._pb_body is None and obj is not None:
                raise ValueError
        except json_format.ParseError:
            # append everything to kwargs that is more powerful in setting attributes
            self._pb_body = self._PbMsg()
            if isinstance(obj, dict):
                kwargs.update(obj)
        except Exception as ex:
            raise ValueError(f'Fail to construct Document from {obj!r}') from ex

        if self._pb_body.id is None or not self._pb_body.id:
            self.id = uuid.uuid1().hex

        if kwargs:
            # check if there are mutually exclusive content fields
            if len(_all_doc_content_keys.intersection(kwargs.keys())) > 1:
                raise ValueError(
                    f'Document content fields are mutually exclusive, please provide only one of {_all_doc_content_keys}'
                )
            self._set_attributes(**kwargs)

    @property
    def weight(self) -> float:
        """
        :return: the weight of the document
        """
        return self._pb_body.weight

    @weight.setter
    def weight(self, value: float):
        """
        Set the weight of the document.

        :param value: the float weight of the document.
        """
        if value is None:
            self.pop('weight')
            return
        self._pb_body.weight = value

    @property
    def modality(self) -> str:
        """
        :return: the modality of the document."""
        return self._pb_body.modality

    @modality.setter
    def modality(self, value: str):
        """Set the modality of the document.

        :param value: The modality of the document
        """
        if value is None:
            self.pop('modality')
            return
        self._pb_body.modality = value

    @property
    def tags(self) -> StructView:
        """Return the `tags` field of this Document as a Python dict

        :return: a Python dict view of the tags.
        """
        return StructView(self._pb_body.tags)

    @tags.setter
    def tags(self, value: Union[Dict, StructView]):
        """Set the `tags` field of this Document to a Python dict

        :param value: a Python dict or a StructView
        """
        if value is None:
            self.pop('tags')
            return

        if isinstance(value, StructView):
            self._pb_body.tags.Clear()
            self._pb_body.tags.update(value._pb_body)
        elif isinstance(value, dict):
            self._pb_body.tags.Clear()
            self._pb_body.tags.update(value)
        else:
            raise TypeError(f'{typename(value)} {value!r} is not supported.')

    @property
    def id(self) -> str:
        """The document id in string.

        :return: the id of this Document
        """
        return self._pb_body.id

    @property
    def parent_id(self) -> str:
        """The document's parent id in string.

        :return: the parent id of this Document
        """
        return self._pb_body.parent_id

    @id.setter
    def id(self, value: str):
        """Set document id to a string value.

        :param value: id as string
        """
        if value is None:
            self.pop('id')
            return
        self._pb_body.id = str(value)

    @parent_id.setter
    def parent_id(self, value: str):
        """Set document's parent id to a string value.

        :param value: id as string
        """
        if value is None:
            self.pop('parent_id')
            return
        self._pb_body.parent_id = str(value)

    @property
    def blob(self) -> 'ArrayType':
        """Return ``blob``, one of the content form of a Document.

        .. note::
            Use :attr:`content` to return the content of a Document

            This property will return the `blob` of the `Document` as a `Dense` or `Sparse` array depending on the actual
            proto instance stored. In the case where the `blob` stored is sparse, it will return them as a `coo` matrix.

        :return: the blob content of thi Document
        """
        return NdArray(self._pb_body.blob).value

    @blob.setter
    def blob(self, value: 'ArrayType'):
        """Set the `blob` to `value`.

        :param value: the array value to set the blob
        """
        if value is None:
            self.pop('blob')
            return
        NdArray(self._pb_body.blob).value = value

    @property
    def embedding(self) -> 'ArrayType':
        """Return ``embedding`` of the content of a Document.

         .. note::
            This property will return the `embedding` of the `Document` as a `Dense` or `Sparse` array depending on the actual
            proto instance stored. In the case where the `embedding` stored is sparse, it will return them as a `coo` matrix.

        :return: the embedding of this Document
        """
        return NdArray(self._pb_body.embedding).value

    @embedding.setter
    def embedding(self, value: 'ArrayType'):
        """Set the ``embedding`` of the content of a Document.

        :param value: the array value to set the embedding
        """
        if value is None:
            self.pop('embedding')
            return
        NdArray(self._pb_body.embedding).value = value

    @property
    @versioned
    def matches(self) -> 'MatchArray':
        """Get all matches of the current document.

        :return: the array of matches attached to this document
        """
        # Problem with cyclic dependency
        from ..array.match import MatchArray

        return MatchArray(self._pb_body.matches, reference_doc=self)

    @matches.setter
    def matches(self, value: Iterable['Document']):
        """Get all chunks of the current document.

        :param value: value to set
        """
        if value is None:
            self.pop('matches')
            return
        self.pop('matches')
        self.matches.extend(value)

    @property
    @versioned
    def chunks(self) -> 'ChunkArray':
        """Get all chunks of the current document.

        :return: the array of chunks of this document
        """
        # Problem with cyclic dependency
        from ..array.chunk import ChunkArray

        return ChunkArray(self._pb_body.chunks, reference_doc=self)

    @chunks.setter
    def chunks(self, value: Iterable['Document']):
        """Get all chunks of the current document.

        :param value: the array of chunks of this document
        """
        if value is None:
            self.pop('chunks')
            return
        self.pop('chunks')
        self.chunks.extend(value)

    @property
    def buffer(self) -> bytes:
        """Return ``buffer``, one of the content form of a Document.

        .. note::
            Use :attr:`content` to return the content of a Document

        :return: the buffer bytes from this document
        """
        return self._pb_body.buffer

    @buffer.setter
    def buffer(self, value: bytes):
        """Set the ``buffer`` to `value`.

        :param value: the bytes value to set the buffer
        """
        if value is None:
            self.pop('buffer')
            return
        self._pb_body.buffer = value

    @property
    def text(self) -> str:
        """Return ``text``, one of the content form of a Document.

        .. note::
            Use :attr:`content` to return the content of a Document

        :return: the text from this document content
        """
        return self._pb_body.text

    @text.setter
    def text(self, value: str):
        """Set the `text` to `value`

        :param value: the text value to set as content
        """
        if value is None:
            self.pop('text')
            self.mime_type = None
            return
        self._pb_body.text = value
        self.mime_type = 'text/plain'

    @property
    def uri(self) -> str:
        """Return the URI of the document.

        :return: the uri of this Document
        """
        return self._pb_body.uri

    @uri.setter
    def uri(self, value: str):
        """Set the URI of the document.

        .. note::
            :attr:`mime_type` will be updated accordingly

        :param value: acceptable URI/URL, raise ``ValueError`` when it is not a valid URI
        """
        if value is None:
            self.pop('uri')
            self.mime_type = None
            return
        self._pb_body.uri = value
        mime_type = mimetypes.guess_type(value)[0]
        if mime_type:
            self.mime_type = mime_type  # Remote http/https contents mime_type will not be recognized.

    @property
    def mime_type(self) -> str:
        """Get MIME type of the document

        :return: the mime_type of this Document
        """
        return self._pb_body.mime_type

    @mime_type.setter
    def mime_type(self, value: str):
        """Set MIME type of the document

        :param value: the acceptable MIME type, raise ``ValueError`` when MIME type is not
                recognizable.
        """
        if value is None:
            self.pop('mime_type')
            return
        if value in _all_mime_types:
            self._pb_body.mime_type = value
        elif value:
            # given but not recognizable, do best guess
            r = mimetypes.guess_type(f'*.{value}')[0]
            if r:
                self._pb_body.mime_type = r
            else:
                self._pb_body.mime_type = value

    @property
    def granularity(self) -> int:
        """Return the granularity of the document.

        :return: the granularity of this Document
        """
        return self._pb_body.granularity

    @granularity.setter
    def granularity(self, value: int):
        """Set the granularity of the document.

        :param value: the value of the granularity to be set
        """
        if value is None:
            self.pop('granularity')
            return
        self._pb_body.granularity = value

    @property
    def adjacency(self) -> int:
        """Return the adjacency of the document.

        :return: the adjacency of this Document
        """
        return self._pb_body.adjacency

    @adjacency.setter
    def adjacency(self, value: int):
        """Set the adjacency of the document.

        :param value: the value of the adjacency to be set
        """
        if value is None:
            self.pop('adjacency')
            return
        self._pb_body.adjacency = value

    @property
    def scores(self):
        """Return the scores of the document.

        :return: the scores attached to this document as `:class:NamedScoreMapping`
        """
        return NamedScoreMap(self._pb_body.scores)

    @scores.setter
    def scores(
        self,
        value: Dict[str, Union['NamedScore', float, np.generic]],
    ):
        """Sets the scores of the `Document`. Specially important to provide the ability to start `scores` as:

            .. highlight:: python
            .. code-block:: python

                from jina import Document
                from jina.types.score import NamedScore
                d = Document(scores={'euclidean': 5, 'cosine': NamedScore(value=0.5)})

        :param value: the dictionary to set the scores
        """
        if value is None:
            self.pop('scores')
            return

        scores = NamedScoreMap(self._pb_body.scores)
        for k, v in value.items():
            scores[k] = v

    @property
    def evaluations(self) -> NamedScoreMap:
        """Return the evaluations of the document.

        :return: the evaluations attached to this document as `:class:NamedScoreMapping`
        """
        return NamedScoreMap(self._pb_body.evaluations)

    @evaluations.setter
    def evaluations(
        self,
        value: Dict[str, Union['NamedScore', float, np.generic]],
    ):
        """Sets the evaluations of the `Document`. Specially important to provide the ability to start `evaluations` as:

            .. highlight:: python
            .. code-block:: python

                from jina import Document
                from jina.types.score import NamedScore
                d = Document(evaluations={'precision': 0.9, 'recall': NamedScore(value=0.5)})

        :param value: the dictionary to set the evaluations
        """
        if value is None:
            self.pop('evaluations')
            return
        scores = NamedScoreMap(self._pb_body.evaluations)
        for k, v in value.items():
            scores[k] = v

    @property
    def location(self) -> Tuple[float]:
        """Get the location information.

        :return: location info in a tuple.
        """
        return tuple(self._pb_body.location)

    @location.setter
    def location(self, value: Sequence[float]):
        """Set the location information of this Document.

        Location could mean the start and end index of a string;
        could be x,y (top, left) coordinate of an image crop; could be timestamp of an audio clip.

        :param value: the location info to be set.
        """
        if value is None:
            self.pop('location')
            return
        self.pop('location')
        self._pb_body.location.extend(value)

    @property
    def offset(self) -> float:
        """Get the offset information of this Document.

        :return: the offset
        """
        return self._pb_body.offset

    @offset.setter
    def offset(self, value: float):
        """Set the offset of this Document

        :param value: the offset value to be set.
        """
        if value is None:
            self.pop('offset')

        self._pb_body.offset = value



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: docarray.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import struct_pb2 as google_dot_protobuf_dot_struct__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
    name='docarray.proto',
    package='docarray',
    syntax='proto3',
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
    serialized_pb=b'\n\x0e\x64ocarray.proto\x12\x08\x64ocarray\x1a\x1cgoogle/protobuf/struct.proto\"A\n\x11\x44\x65nseNdArrayProto\x12\x0e\n\x06\x62uffer\x18\x01 \x01(\x0c\x12\r\n\x05shape\x18\x02 \x03(\r\x12\r\n\x05\x64type\x18\x03 \x01(\t\"\xb6\x01\n\x0cNdArrayProto\x12,\n\x05\x64\x65nse\x18\x01 \x01(\x0b\x32\x1b.docarray.DenseNdArrayProtoH\x00\x12.\n\x06sparse\x18\x02 \x01(\x0b\x32\x1c.docarray.SparseNdArrayProtoH\x00\x12\x10\n\x08\x63ls_name\x18\x03 \x01(\t\x12+\n\nparameters\x18\x04 \x01(\x0b\x32\x17.google.protobuf.StructB\t\n\x07\x63ontent\"~\n\x12SparseNdArrayProto\x12,\n\x07indices\x18\x01 \x01(\x0b\x32\x1b.docarray.DenseNdArrayProto\x12+\n\x06values\x18\x02 \x01(\x0b\x32\x1b.docarray.DenseNdArrayProto\x12\r\n\x05shape\x18\x03 \x03(\r\"\x83\x01\n\x0fNamedScoreProto\x12\r\n\x05value\x18\x01 \x01(\x02\x12\x0f\n\x07op_name\x18\x02 \x01(\t\x12\x13\n\x0b\x64\x65scription\x18\x03 \x01(\t\x12+\n\x08operands\x18\x04 \x03(\x0b\x32\x19.docarray.NamedScoreProto\x12\x0e\n\x06ref_id\x18\x05 \x01(\t\"{\n\nGraphProto\x12)\n\tadjacency\x18\x01 \x01(\x0b\x32\x16.docarray.NdArrayProto\x12.\n\redge_features\x18\x02 \x01(\x0b\x32\x17.google.protobuf.Struct\x12\x12\n\nundirected\x18\x03 \x01(\x08\"\xc1\x05\n\rDocumentProto\x12\n\n\x02id\x18\x01 \x01(\t\x12\x13\n\x0bgranularity\x18\x0e \x01(\r\x12\x11\n\tadjacency\x18\x16 \x01(\r\x12\x11\n\tparent_id\x18\x10 \x01(\t\x12\x10\n\x06\x62uffer\x18\x03 \x01(\x0cH\x00\x12&\n\x04\x62lob\x18\x0c \x01(\x0b\x32\x16.docarray.NdArrayProtoH\x00\x12\x0e\n\x04text\x18\r \x01(\tH\x00\x12\'\n\x06\x63hunks\x18\x04 \x03(\x0b\x32\x17.docarray.DocumentProto\x12\x0e\n\x06weight\x18\x05 \x01(\x02\x12(\n\x07matches\x18\x08 \x03(\x0b\x32\x17.docarray.DocumentProto\x12\x0b\n\x03uri\x18\t \x01(\t\x12\x11\n\tmime_type\x18\n \x01(\t\x12%\n\x04tags\x18\x0b \x01(\x0b\x32\x17.google.protobuf.Struct\x12\x10\n\x08location\x18\x11 \x03(\x02\x12\x0e\n\x06offset\x18\x12 \x01(\x02\x12)\n\tembedding\x18\x13 \x01(\x0b\x32\x16.docarray.NdArrayProto\x12\x33\n\x06scores\x18\x1c \x03(\x0b\x32#.docarray.DocumentProto.ScoresEntry\x12\x10\n\x08modality\x18\x15 \x01(\t\x12=\n\x0b\x65valuations\x18\x1d \x03(\x0b\x32(.docarray.DocumentProto.EvaluationsEntry\x1aH\n\x0bScoresEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12(\n\x05value\x18\x02 \x01(\x0b\x32\x19.docarray.NamedScoreProto:\x02\x38\x01\x1aM\n\x10\x45valuationsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12(\n\x05value\x18\x02 \x01(\x0b\x32\x19.docarray.NamedScoreProto:\x02\x38\x01\x42\t\n\x07\x63ontent\";\n\x12\x44ocumentArrayProto\x12%\n\x04\x64ocs\x18\x01 \x03(\x0b\x32\x17.docarray.DocumentProtob\x06proto3',
    dependencies=[
        google_dot_protobuf_dot_struct__pb2.DESCRIPTOR,
    ],
)


_DENSENDARRAYPROTO = _descriptor.Descriptor(
    name='DenseNdArrayProto',
    full_name='docarray.DenseNdArrayProto',
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name='buffer',
            full_name='docarray.DenseNdArrayProto.buffer',
            index=0,
            number=1,
            type=12,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"",
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name='shape',
            full_name='docarray.DenseNdArrayProto.shape',
            index=1,
            number=2,
            type=13,
            cpp_type=3,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name='dtype',
            full_name='docarray.DenseNdArrayProto.dtype',
            index=2,
            number=3,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode('utf-8'),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax='proto3',
    extension_ranges=[],
    oneofs=[],
    serialized_start=58,
    serialized_end=123,
)


_NDARRAYPROTO = _descriptor.Descriptor(
    name='NdArrayProto',
    full_name='docarray.NdArrayProto',
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name='dense',
            full_name='docarray.NdArrayProto.dense',
            index=0,
            number=1,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name='sparse',
            full_name='docarray.NdArrayProto.sparse',
            index=1,
            number=2,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name='cls_name',
            full_name='docarray.NdArrayProto.cls_name',
            index=2,
            number=3,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode('utf-8'),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name='parameters',
            full_name='docarray.NdArrayProto.parameters',
            index=3,
            number=4,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax='proto3',
    extension_ranges=[],
    oneofs=[
        _descriptor.OneofDescriptor(
            name='content',
            full_name='docarray.NdArrayProto.content',
            index=0,
            containing_type=None,
            create_key=_descriptor._internal_create_key,
            fields=[],
        ),
    ],
    serialized_start=126,
    serialized_end=308,
)


_SPARSENDARRAYPROTO = _descriptor.Descriptor(
    name='SparseNdArrayProto',
    full_name='docarray.SparseNdArrayProto',
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name='indices',
            full_name='docarray.SparseNdArrayProto.indices',
            index=0,
            number=1,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name='values',
            full_name='docarray.SparseNdArrayProto.values',
            index=1,
            number=2,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name='shape',
            full_name='docarray.SparseNdArrayProto.shape',
            index=2,
            number=3,
            type=13,
            cpp_type=3,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax='proto3',
    extension_ranges=[],
    oneofs=[],
    serialized_start=310,
    serialized_end=436,
)


_NAMEDSCOREPROTO = _descriptor.Descriptor(
    name='NamedScoreProto',
    full_name='docarray.NamedScoreProto',
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name='value',
            full_name='docarray.NamedScoreProto.value',
            index=0,
            number=1,
            type=2,
            cpp_type=6,
            label=1,
            has_default_value=False,
            default_value=float(0),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name='op_name',
            full_name='docarray.NamedScoreProto.op_name',
            index=1,
            number=2,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode('utf-8'),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name='description',
            full_name='docarray.NamedScoreProto.description',
            index=2,
            number=3,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode('utf-8'),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name='operands',
            full_name='docarray.NamedScoreProto.operands',
            index=3,
            number=4,
            type=11,
            cpp_type=10,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name='ref_id',
            full_name='docarray.NamedScoreProto.ref_id',
            index=4,
            number=5,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode('utf-8'),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax='proto3',
    extension_ranges=[],
    oneofs=[],
    serialized_start=439,
    serialized_end=570,
)


_GRAPHPROTO = _descriptor.Descriptor(
    name='GraphProto',
    full_name='docarray.GraphProto',
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name='adjacency',
            full_name='docarray.GraphProto.adjacency',
            index=0,
            number=1,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name='edge_features',
            full_name='docarray.GraphProto.edge_features',
            index=1,
            number=2,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name='undirected',
            full_name='docarray.GraphProto.undirected',
            index=2,
            number=3,
            type=8,
            cpp_type=7,
            label=1,
            has_default_value=False,
            default_value=False,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax='proto3',
    extension_ranges=[],
    oneofs=[],
    serialized_start=572,
    serialized_end=695,
)


_DOCUMENTPROTO_SCORESENTRY = _descriptor.Descriptor(
    name='ScoresEntry',
    full_name='docarray.DocumentProto.ScoresEntry',
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name='key',
            full_name='docarray.DocumentProto.ScoresEntry.key',
            index=0,
            number=1,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode('utf-8'),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name='value',
            full_name='docarray.DocumentProto.ScoresEntry.value',
            index=1,
            number=2,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=b'8\001',
    is_extendable=False,
    syntax='proto3',
    extension_ranges=[],
    oneofs=[],
    serialized_start=1241,
    serialized_end=1313,
)

_DOCUMENTPROTO_EVALUATIONSENTRY = _descriptor.Descriptor(
    name='EvaluationsEntry',
    full_name='docarray.DocumentProto.EvaluationsEntry',
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name='key',
            full_name='docarray.DocumentProto.EvaluationsEntry.key',
            index=0,
            number=1,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode('utf-8'),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name='value',
            full_name='docarray.DocumentProto.EvaluationsEntry.value',
            index=1,
            number=2,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=b'8\001',
    is_extendable=False,
    syntax='proto3',
    extension_ranges=[],
    oneofs=[],
    serialized_start=1315,
    serialized_end=1392,
)

_DOCUMENTPROTO = _descriptor.Descriptor(
    name='DocumentProto',
    full_name='docarray.DocumentProto',
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name='id',
            full_name='docarray.DocumentProto.id',
            index=0,
            number=1,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode('utf-8'),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name='granularity',
            full_name='docarray.DocumentProto.granularity',
            index=1,
            number=14,
            type=13,
            cpp_type=3,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name='adjacency',
            full_name='docarray.DocumentProto.adjacency',
            index=2,
            number=22,
            type=13,
            cpp_type=3,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name='parent_id',
            full_name='docarray.DocumentProto.parent_id',
            index=3,
            number=16,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode('utf-8'),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name='buffer',
            full_name='docarray.DocumentProto.buffer',
            index=4,
            number=3,
            type=12,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"",
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name='blob',
            full_name='docarray.DocumentProto.blob',
            index=5,
            number=12,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name='text',
            full_name='docarray.DocumentProto.text',
            index=6,
            number=13,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode('utf-8'),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name='chunks',
            full_name='docarray.DocumentProto.chunks',
            index=7,
            number=4,
            type=11,
            cpp_type=10,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name='weight',
            full_name='docarray.DocumentProto.weight',
            index=8,
            number=5,
            type=2,
            cpp_type=6,
            label=1,
            has_default_value=False,
            default_value=float(0),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name='matches',
            full_name='docarray.DocumentProto.matches',
            index=9,
            number=8,
            type=11,
            cpp_type=10,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name='uri',
            full_name='docarray.DocumentProto.uri',
            index=10,
            number=9,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode('utf-8'),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name='mime_type',
            full_name='docarray.DocumentProto.mime_type',
            index=11,
            number=10,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode('utf-8'),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name='tags',
            full_name='docarray.DocumentProto.tags',
            index=12,
            number=11,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name='location',
            full_name='docarray.DocumentProto.location',
            index=13,
            number=17,
            type=2,
            cpp_type=6,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name='offset',
            full_name='docarray.DocumentProto.offset',
            index=14,
            number=18,
            type=2,
            cpp_type=6,
            label=1,
            has_default_value=False,
            default_value=float(0),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name='embedding',
            full_name='docarray.DocumentProto.embedding',
            index=15,
            number=19,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name='scores',
            full_name='docarray.DocumentProto.scores',
            index=16,
            number=28,
            type=11,
            cpp_type=10,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name='modality',
            full_name='docarray.DocumentProto.modality',
            index=17,
            number=21,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode('utf-8'),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name='evaluations',
            full_name='docarray.DocumentProto.evaluations',
            index=18,
            number=29,
            type=11,
            cpp_type=10,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[
        _DOCUMENTPROTO_SCORESENTRY,
        _DOCUMENTPROTO_EVALUATIONSENTRY,
    ],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax='proto3',
    extension_ranges=[],
    oneofs=[
        _descriptor.OneofDescriptor(
            name='content',
            full_name='docarray.DocumentProto.content',
            index=0,
            containing_type=None,
            create_key=_descriptor._internal_create_key,
            fields=[],
        ),
    ],
    serialized_start=698,
    serialized_end=1403,
)


_DOCUMENTARRAYPROTO = _descriptor.Descriptor(
    name='DocumentArrayProto',
    full_name='docarray.DocumentArrayProto',
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name='docs',
            full_name='docarray.DocumentArrayProto.docs',
            index=0,
            number=1,
            type=11,
            cpp_type=10,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax='proto3',
    extension_ranges=[],
    oneofs=[],
    serialized_start=1405,
    serialized_end=1464,
)

_NDARRAYPROTO.fields_by_name['dense'].message_type = _DENSENDARRAYPROTO
_NDARRAYPROTO.fields_by_name['sparse'].message_type = _SPARSENDARRAYPROTO
_NDARRAYPROTO.fields_by_name[
    'parameters'
].message_type = google_dot_protobuf_dot_struct__pb2._STRUCT
_NDARRAYPROTO.oneofs_by_name['content'].fields.append(
    _NDARRAYPROTO.fields_by_name['dense']
)
_NDARRAYPROTO.fields_by_name['dense'].containing_oneof = _NDARRAYPROTO.oneofs_by_name[
    'content'
]
_NDARRAYPROTO.oneofs_by_name['content'].fields.append(
    _NDARRAYPROTO.fields_by_name['sparse']
)
_NDARRAYPROTO.fields_by_name['sparse'].containing_oneof = _NDARRAYPROTO.oneofs_by_name[
    'content'
]
_SPARSENDARRAYPROTO.fields_by_name['indices'].message_type = _DENSENDARRAYPROTO
_SPARSENDARRAYPROTO.fields_by_name['values'].message_type = _DENSENDARRAYPROTO
_NAMEDSCOREPROTO.fields_by_name['operands'].message_type = _NAMEDSCOREPROTO
_GRAPHPROTO.fields_by_name['adjacency'].message_type = _NDARRAYPROTO
_GRAPHPROTO.fields_by_name[
    'edge_features'
].message_type = google_dot_protobuf_dot_struct__pb2._STRUCT
_DOCUMENTPROTO_SCORESENTRY.fields_by_name['value'].message_type = _NAMEDSCOREPROTO
_DOCUMENTPROTO_SCORESENTRY.containing_type = _DOCUMENTPROTO
_DOCUMENTPROTO_EVALUATIONSENTRY.fields_by_name['value'].message_type = _NAMEDSCOREPROTO
_DOCUMENTPROTO_EVALUATIONSENTRY.containing_type = _DOCUMENTPROTO
_DOCUMENTPROTO.fields_by_name['blob'].message_type = _NDARRAYPROTO
_DOCUMENTPROTO.fields_by_name['chunks'].message_type = _DOCUMENTPROTO
_DOCUMENTPROTO.fields_by_name['matches'].message_type = _DOCUMENTPROTO
_DOCUMENTPROTO.fields_by_name[
    'tags'
].message_type = google_dot_protobuf_dot_struct__pb2._STRUCT
_DOCUMENTPROTO.fields_by_name['embedding'].message_type = _NDARRAYPROTO
_DOCUMENTPROTO.fields_by_name['scores'].message_type = _DOCUMENTPROTO_SCORESENTRY
_DOCUMENTPROTO.fields_by_name[
    'evaluations'
].message_type = _DOCUMENTPROTO_EVALUATIONSENTRY
_DOCUMENTPROTO.oneofs_by_name['content'].fields.append(
    _DOCUMENTPROTO.fields_by_name['buffer']
)
_DOCUMENTPROTO.fields_by_name[
    'buffer'
].containing_oneof = _DOCUMENTPROTO.oneofs_by_name['content']
_DOCUMENTPROTO.oneofs_by_name['content'].fields.append(
    _DOCUMENTPROTO.fields_by_name['blob']
)
_DOCUMENTPROTO.fields_by_name['blob'].containing_oneof = _DOCUMENTPROTO.oneofs_by_name[
    'content'
]
_DOCUMENTPROTO.oneofs_by_name['content'].fields.append(
    _DOCUMENTPROTO.fields_by_name['text']
)
_DOCUMENTPROTO.fields_by_name['text'].containing_oneof = _DOCUMENTPROTO.oneofs_by_name[
    'content'
]
_DOCUMENTARRAYPROTO.fields_by_name['docs'].message_type = _DOCUMENTPROTO
DESCRIPTOR.message_types_by_name['DenseNdArrayProto'] = _DENSENDARRAYPROTO
DESCRIPTOR.message_types_by_name['NdArrayProto'] = _NDARRAYPROTO
DESCRIPTOR.message_types_by_name['SparseNdArrayProto'] = _SPARSENDARRAYPROTO
DESCRIPTOR.message_types_by_name['NamedScoreProto'] = _NAMEDSCOREPROTO
DESCRIPTOR.message_types_by_name['GraphProto'] = _GRAPHPROTO
DESCRIPTOR.message_types_by_name['DocumentProto'] = _DOCUMENTPROTO
DESCRIPTOR.message_types_by_name['DocumentArrayProto'] = _DOCUMENTARRAYPROTO
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

DenseNdArrayProto = _reflection.GeneratedProtocolMessageType(
    'DenseNdArrayProto',
    (_message.Message,),
    {
        'DESCRIPTOR': _DENSENDARRAYPROTO,
        '__module__': 'docarray_pb2'
        # @@protoc_insertion_point(class_scope:docarray.DenseNdArrayProto)
    },
)
_sym_db.RegisterMessage(DenseNdArrayProto)

NdArrayProto = _reflection.GeneratedProtocolMessageType(
    'NdArrayProto',
    (_message.Message,),
    {
        'DESCRIPTOR': _NDARRAYPROTO,
        '__module__': 'docarray_pb2'
        # @@protoc_insertion_point(class_scope:docarray.NdArrayProto)
    },
)
_sym_db.RegisterMessage(NdArrayProto)

SparseNdArrayProto = _reflection.GeneratedProtocolMessageType(
    'SparseNdArrayProto',
    (_message.Message,),
    {
        'DESCRIPTOR': _SPARSENDARRAYPROTO,
        '__module__': 'docarray_pb2'
        # @@protoc_insertion_point(class_scope:docarray.SparseNdArrayProto)
    },
)
_sym_db.RegisterMessage(SparseNdArrayProto)

NamedScoreProto = _reflection.GeneratedProtocolMessageType(
    'NamedScoreProto',
    (_message.Message,),
    {
        'DESCRIPTOR': _NAMEDSCOREPROTO,
        '__module__': 'docarray_pb2'
        # @@protoc_insertion_point(class_scope:docarray.NamedScoreProto)
    },
)
_sym_db.RegisterMessage(NamedScoreProto)

GraphProto = _reflection.GeneratedProtocolMessageType(
    'GraphProto',
    (_message.Message,),
    {
        'DESCRIPTOR': _GRAPHPROTO,
        '__module__': 'docarray_pb2'
        # @@protoc_insertion_point(class_scope:docarray.GraphProto)
    },
)
_sym_db.RegisterMessage(GraphProto)

DocumentProto = _reflection.GeneratedProtocolMessageType(
    'DocumentProto',
    (_message.Message,),
    {
        'ScoresEntry': _reflection.GeneratedProtocolMessageType(
            'ScoresEntry',
            (_message.Message,),
            {
                'DESCRIPTOR': _DOCUMENTPROTO_SCORESENTRY,
                '__module__': 'docarray_pb2'
                # @@protoc_insertion_point(class_scope:docarray.DocumentProto.ScoresEntry)
            },
        ),
        'EvaluationsEntry': _reflection.GeneratedProtocolMessageType(
            'EvaluationsEntry',
            (_message.Message,),
            {
                'DESCRIPTOR': _DOCUMENTPROTO_EVALUATIONSENTRY,
                '__module__': 'docarray_pb2'
                # @@protoc_insertion_point(class_scope:docarray.DocumentProto.EvaluationsEntry)
            },
        ),
        'DESCRIPTOR': _DOCUMENTPROTO,
        '__module__': 'docarray_pb2'
        # @@protoc_insertion_point(class_scope:docarray.DocumentProto)
    },
)
_sym_db.RegisterMessage(DocumentProto)
_sym_db.RegisterMessage(DocumentProto.ScoresEntry)
_sym_db.RegisterMessage(DocumentProto.EvaluationsEntry)

DocumentArrayProto = _reflection.GeneratedProtocolMessageType(
    'DocumentArrayProto',
    (_message.Message,),
    {
        'DESCRIPTOR': _DOCUMENTARRAYPROTO,
        '__module__': 'docarray_pb2'
        # @@protoc_insertion_point(class_scope:docarray.DocumentArrayProto)
    },
)
_sym_db.RegisterMessage(DocumentArrayProto)


_DOCUMENTPROTO_SCORESENTRY._options = None
_DOCUMENTPROTO_EVALUATIONSENTRY._options = None
# @@protoc_insertion_point(module_scope)



import copy
import json
import os.path
import shutil
import tempfile
import threading
import warnings
from math import sqrt, ceil, floor
from typing import Optional

import numpy as np


from ...helper import random_port, __resources_path__


class PlotMixin:
    """Helper functions for plotting the arrays. """

    def plot_embeddings(
        self,
        title: str = 'MyDocumentArray',
        path: Optional[str] = None,
        image_sprites: bool = False,
        min_image_size: int = 16,
        channel_axis: int = -1,
        start_server: bool = True,
        port: Optional[int] = None,
    ) -> str:
        """Interactively visualize :attr:`.embeddings` using the Embedding Projector.

        :param title: the title of this visualization. If you want to compare multiple embeddings at the same time,
                make sure to give different names each time and set ``path`` to the same value.
        :param port: if set, run the embedding-projector frontend at given port. Otherwise a random port is used.
        :param image_sprites: if set, visualize the dots using :attr:`.uri` and :attr:`.blob`.
        :param path: if set, then append the visualization to an existing folder, where you can compare multiple
            embeddings at the same time. Make sure to use a different ``title`` each time .
        :param min_image_size: only used when `image_sprites=True`. the minimum size of the image
        :param channel_axis: only used when `image_sprites=True`. the axis id of the color channel, ``-1`` indicates the color channel info at the last axis
        :param start_server: if set, start a HTTP server and open the frontend directly. Otherwise, you need to rely on ``return`` path and serve by yourself.
        :return: the path to the embeddings visualization info.
        """
        path = path or tempfile.mkdtemp()
        emb_fn = f'{title}.tsv'
        meta_fn = f'{title}.metas.tsv'
        config_fn = f'config.json'
        sprite_fn = f'{title}.png'

        if image_sprites:
            img_per_row = ceil(sqrt(len(self)))
            canvas_size = min(img_per_row * min_image_size, 8192)
            img_size = max(int(canvas_size / img_per_row), min_image_size)

            max_docs = ceil(canvas_size / img_size) ** 2
            if len(self) > max_docs:
                warnings.warn(
                    f'''
                    {self!r} has more than {max_docs} elements, which is the maximum number of image sprites can support. 
                    The resulting visualization may not be correct. You can do the following:
                    
                    - use fewer images: `da[:10000].plot_embeddings()`
                    - reduce the `min_image_size` to a smaller number, say 8 or 4 (but bear in mind you can hardly recognize anything with a 4x4 image)
                    - turn off `image_sprites` via `da.plot_embeddings(image_sprites=False)`
                    '''
                )

            self.plot_image_sprites(
                os.path.join(path, sprite_fn),
                canvas_size=canvas_size,
                min_size=min_image_size,
                channel_axis=channel_axis,
            )

        self.save_embeddings_csv(os.path.join(path, emb_fn), delimiter='\t')
        self.save_csv(
            os.path.join(path, meta_fn),
            exclude_fields=('embedding', 'blob', 'scores'),
            dialect='excel-tab',
        )

        _epj_config = {
            'embeddings': [
                {
                    'tensorName': title,
                    'tensorShape': list(self.embeddings.shape),
                    'tensorPath': f'/static/{emb_fn}',
                    'metadataPath': f'/static/{meta_fn}',
                    'sprite': {
                        'imagePath': f'/static/{sprite_fn}',
                        'singleImageDim': (img_size,) * 2,
                    }
                    if image_sprites
                    else {},
                }
            ]
        }

        if os.path.exists(os.path.join(path, config_fn)):
            with open(os.path.join(path, config_fn)) as fp:
                old_config = json.load(fp)
                _epj_config['embeddings'].extend(old_config.get('embeddings', []))

        with open(os.path.join(path, config_fn), 'w') as fp:
            json.dump(_epj_config, fp)

        shutil.copyfile(
            os.path.join(__resources_path__, 'embedding-projector/index.html'),
            os.path.join(path, 'index.html'),
        )

        if start_server:

            def _get_fastapi_app():
                from fastapi import FastAPI
                from starlette.middleware.cors import CORSMiddleware
                from starlette.staticfiles import StaticFiles

                app = FastAPI()
                app.add_middleware(
                    CORSMiddleware,
                    allow_origins=['*'],
                    allow_credentials=True,
                    allow_methods=['*'],
                    allow_headers=['*'],
                )
                app.mount('/static', StaticFiles(directory=path), name='static')
                return app

            import uvicorn

            app = _get_fastapi_app()
            port = port or random_port()
            t_m = threading.Thread(
                target=uvicorn.run,
                kwargs=dict(app=app, port=port, log_level='error'),
                daemon=True,
            )
            t_m.start()
            url_html_path = (
                f'http://localhost:{port}/static/index.html?config={config_fn}'
            )
            try:
                import webbrowser

                webbrowser.open(url_html_path, new=2)
            except:
                pass  # intentional pass, browser support isn't cross-platform
            finally:
                print(
                    f'You should see a webpage opened in your browser, '
                    f'if not, you may open {url_html_path} manually'
                )
            t_m.join()
        return path

    def plot_embeddings_legacy(
        self,
        output: Optional[str] = None,
        title: Optional[str] = None,
        colored_attr: Optional[str] = None,
        colormap: str = 'rainbow',
        method: str = 'pca',
        show_axis: bool = False,
        **kwargs,
    ):
        """Plot embeddings in a 2D projection with the PCA algorithm. This function requires ``matplotlib`` installed.

        If `tag_name` is provided the plot uses a distinct color for each unique tag value in the
        documents of the DocumentArray.

        :param output: Optional path to store the visualization. If not given, show in UI
        :param title: Optional title of the plot. When not given, the default title is used.
        :param colored_attr: Optional str that specifies attribute used to color the plot, it supports dunder expression
            such as `tags__label`, `matches__0__id`.
        :param colormap: the colormap string supported by matplotlib.
        :param method: the visualization method, available `pca`, `tsne`. `pca` is fast but may not well represent
                nonlinear relationship of high-dimensional data. `tsne` requires scikit-learn to be installed and is
                much slower.
        :param show_axis: If set, axis and bounding box of the plot will be printed.
        :param kwargs: extra kwargs pass to matplotlib.plot
        """

        x_mat = self.embeddings
        assert isinstance(
            x_mat, np.ndarray
        ), f'Type {type(x_mat)} not currently supported, use np.ndarray embeddings'

        if method == 'tsne':
            from sklearn.manifold import TSNE

            x_mat_2d = TSNE(n_components=2).fit_transform(x_mat)
        else:
            from ...math.dimensionality_reduction import PCA

            x_mat_2d = PCA(n_components=2).fit_transform(x_mat)

        plt_kwargs = {
            'x': x_mat_2d[:, 0],
            'y': x_mat_2d[:, 1],
            'alpha': 0.2,
            'marker': '.',
        }

        import matplotlib.pyplot as plt

        plt.figure(figsize=(8, 8))

        plt.title(title or f'{len(x_mat)} Documents with {method}')

        if colored_attr:
            tags = []

            for x in self:
                try:
                    tags.append(getattr(x, colored_attr))
                except (KeyError, AttributeError, ValueError):
                    tags.append(None)
            tag_to_num = {tag: num for num, tag in enumerate(set(tags))}
            plt_kwargs['c'] = np.array([tag_to_num[ni] for ni in tags])
            plt_kwargs['cmap'] = plt.get_cmap(colormap)

        # update the plt_kwargs
        plt_kwargs.update(kwargs)

        plt.scatter(**plt_kwargs)

        if not show_axis:
            plt.gca().set_axis_off()
            plt.gca().xaxis.set_major_locator(plt.NullLocator())
            plt.gca().yaxis.set_major_locator(plt.NullLocator())

        if output:
            plt.savefig(output, bbox_inches='tight', pad_inches=0.1)
        else:
            plt.show()

    def plot_image_sprites(
        self,
        output: Optional[str] = None,
        canvas_size: int = 512,
        min_size: int = 16,
        channel_axis: int = -1,
    ) -> None:
        """Generate a sprite image for all image blobs in this DocumentArray-like object.

        An image sprite is a collection of images put into a single image. It is always square-sized.
        Each sub-image is also square-sized and equally-sized.

        :param output: Optional path to store the visualization. If not given, show in UI
        :param canvas_size: the size of the canvas
        :param min_size: the minimum size of the image
        :param channel_axis: the axis id of the color channel, ``-1`` indicates the color channel info at the last axis
        """
        if not self:
            raise ValueError(f'{self!r} is empty')

        import matplotlib.pyplot as plt

        img_per_row = ceil(sqrt(len(self)))
        img_size = int(canvas_size / img_per_row)

        if img_size < min_size:
            # image is too small, recompute the size
            img_size = min_size
            img_per_row = int(canvas_size / img_size)

        max_num_img = img_per_row ** 2
        sprite_img = np.zeros(
            [img_size * img_per_row, img_size * img_per_row, 3], dtype='uint8'
        )
        img_id = 0
        for d in self:
            _d = copy.deepcopy(d)
            if _d.content_type != 'blob':
                _d.load_uri_to_image_blob()
                channel_axis = -1

            _d.set_image_blob_channel_axis(channel_axis, -1).set_image_blob_shape(
                shape=(img_size, img_size)
            )

            row_id = floor(img_id / img_per_row)
            col_id = img_id % img_per_row
            sprite_img[
                (row_id * img_size) : ((row_id + 1) * img_size),
                (col_id * img_size) : ((col_id + 1) * img_size),
            ] = _d.blob

            img_id += 1
            if img_id >= max_num_img:
                break

        from PIL import Image

        im = Image.fromarray(sprite_img)

        if output:
            with open(output, 'wb') as fp:
                im.save(fp)
        else:
            plt.gca().set_axis_off()
            plt.subplots_adjust(top=1, bottom=0, right=1, left=0, hspace=0, wspace=0)
            plt.margins(0, 0)
            plt.gca().xaxis.set_major_locator(plt.NullLocator())
            plt.gca().yaxis.set_major_locator(plt.NullLocator())
            plt.imshow(im)
            plt.show()
